[BITS 32]

FDC_DOR	equ	03F2h	;Digital Output Register
FDC_MSR	equ	03F4h	;Main Status Register(input)
FDC_DSR	equ	03F4h	;DataRate Select Register(output)
FDC_DATA	equ	03F5h	;Data Register
FDC_DIR	equ	03F7h	;Digital Input Register(input)
FDC_CCR	equ	03F7h	;Configuration Control Register(output)

CMD_SPECIFY	equ	03h	;Specify Drive Timings
CMD_WRITE	equ	0C5h	;Write Data(+ MT,MFM)
CMD_READ	equ	0E6h	;Read Data(+ MT,MFM,SK)
CMD_CALIBRATE equ	07h	;Recalibrate drive
CMD_SENSEINT equ	08h	;Sense Interrupt status
CMD_FORMAT	equ	4Dh	;Format Track(+ MFM)
CMD_SEEK	equ	0Fh	;Seek Track
CMD_VERSION	equ	10h	;FDC Version

DISK144_TRACKS	equ	80
DISK144_HEADS	equ	2
DISK144_SPT		equ	18
DISK144_GAP3RW	equ	1Bh
DISK144_GAP3FMT	equ	54h

;FDC_Motor_Delay	equ	18*8	; 2 seconds
FDC_Motor_Delay	equ	18*2	; 500 ms

FDC_IRQ_Handler:
	cli
	pushad

	mov dword[FDC_IRQ_Done],1

	mov al,20h	; Send EOI to PIC
	out 20h,al	;

	popad
	iret

FDC_WaitIRQ:
	push ebp
	mov ebp,esp

	xor ecx,ecx
	mov edi,status
@fdc_wait:
	cmp dword[FDC_IRQ_Done],0
	je @fdc_wait

@fdc_w1:
	mov edx,FDC_MSR
	xor eax,eax
	in al,dx
	and al,08h
	jz @fdc_w2

	call FDC_ReadByte
	mov [edi+ecx],al
	inc ecx
	cmp ecx,7
	jne @fdc_w1

@fdc_w2:
	mov [FDC_IRQ_Done],dword 0
	mov esp,ebp
	pop ebp
	ret

FDC_SenseInt:
	push ebp
	mov ebp,esp

	push dword CMD_SENSEINT
	call FDC_WriteByte

	call FDC_ReadByte
	mov byte[sr0],al

	call FDC_ReadByte
	mov byte[fdc_track],al

	mov esp,ebp
	pop ebp
	ret

FDC_Reset:
	push ebp
	mov ebp,esp

	call FDC_MotorOff

	mov dx,FDC_DSR	; Set data rate to 500 kBit/s
	xor eax,eax
	out dx,al

	mov dx,FDC_DOR
	mov eax,0Ch
	out dx,al		; Reenable Interrupts

	mov [FDC_IRQ_Done],dword 1
	call FDC_WaitIRQ
	call FDC_SenseInt

	push dword CMD_SPECIFY
	call FDC_WriteByte

	push dword 0DFh	; SRT = 3ms, HUT = 240ms
	call FDC_WriteByte; HLT = 16ms, ND = 0 

	push dword 02h
	call FDC_WriteByte

	push dword 1
	call FDC_Seek
	call FDC_Calibrate
	mov esp,ebp
	pop ebp
	ret

FDC_MotorOn:
	push ebp
	mov ebp,esp

	cmp dword[FDC_MotorOnFlag],0
	je @fmon_1

	mov dx,FDC_DOR
	mov al,01Ch
	out dx,al

	mov dword[FDC_Motor_Timer],FDC_Motor_Delay
@fmon_1:
	mov esp,ebp
	pop ebp
	ret

FDC_MotorOff:
	push ebp
	mov ebp,esp

	mov edx,FDC_DOR
	xor eax,eax
	out dx,al

	xor eax,eax
	mov dword[FDC_MotorOnFlag],eax
	mov dword[FDC_Motor_Timer],eax

	mov esp,ebp
	pop ebp
	ret

FDC_Calibrate:
	push ebp
	mov ebp,esp

	call FDC_MotorOn

	push dword CMD_CALIBRATE
	call FDC_WriteByte

	push dword 0
	call FDC_WriteByte

	call FDC_WaitIRQ
	call FDC_SenseInt

	call FDC_MotorOff
	mov esp,ebp
	pop ebp
	ret

; Send in track as param
FDC_Seek:
	push ebp
	mov ebp,esp

	mov eax,[ebp+8]
	cmp al,[fdc_track]
	je @fds_true
	call FDC_MotorOn

	push dword CMD_SEEK
	call FDC_WriteByte

	push dword 0
	call FDC_WriteByte

	mov eax,[ebp+8]
	push eax
	call FDC_WriteByte

	mov dword[FDC_Timer],10

@fds1:cmp dword[FDC_Timer],0
	jne @fds1

	call FDC_MotorOff

	cmp byte[sr0],20h
	jne @fds_false

	mov eax,[ebp+8]
	cmp al,[fdc_track]
	jne @fds_false

@fds_true:
	mov eax,1
	jmp @fds_exit

@fds_false:
	xor eax,eax
@fds_exit:
	mov esp,ebp
	pop ebp
	ret 4

; Params
; FDC_ReadWrite(int lba,read/write)
; read/write = 0 means read
; read/write = 1 means write
; Reads block into 7000h
FDC_ReadWrite:
	push ebp
	mov ebp,esp

	mov dword[temp],3
@fdcrw_loop:
	mov eax,[ebp+8h]	; get block address
	xor edx,edx
	mov ecx,DISK144_SPT
	div ecx
	inc edx	; DX is now the sector no.
	mov [ebp+10h],edx	; save it
	xor edx,edx
	mov ebx,DISK144_HEADS
	div ebx	; EAX = track no. -- EDX = head

	mov [ebp+14h],edx	; save them all on stack
	mov [ebp+18h],eax	; save on stack

	call FDC_MotorOn
	mov dword[FDC_Timer],18*2
@fdcrw_1:
	cmp dword[FDC_Timer],0
	jne @fdcrw_1

	mov eax,[ebp+18h]
	push eax
	call FDC_Seek
	or eax,eax
	je @fdcrw_f1
	jmp @fdcrw_2
@fdcrw_f1:
	jmp @fdcrw_false
@fdcrw_2:
	mov edx,FDC_DSR	; set speed to 500 kbit/s
	xor eax,eax
	out dx,al

	mov eax,[ebp+0Ch]
	or eax,eax
	je @fdcrw_read

	call DMA_Setup_Write
	push dword CMD_WRITE
	call FDC_WriteByte
	jmp @fdcrw_continue
@fdcrw_read:
	call DMA_Setup_Read
	push dword CMD_READ
	call FDC_WriteByte
@fdcrw_continue:
	mov ebx,[ebp+14h]	; head
	mov eax,ebx
	shl eax,2

	push eax		; send head << 2
	call FDC_WriteByte

	mov eax,[ebp+18h]	; send track
	call FDC_WriteByte

	push ebx		; send head
	call FDC_WriteByte

	mov eax,[ebp+10h]	; send sector
	push eax
	call FDC_WriteByte

	push dword 2
	call FDC_WriteByte	; 512 bytes/sector

	mov eax,DISK144_SPT
	push eax
	call FDC_WriteByte	; Send sectors per track

	mov eax,DISK144_GAP3RW	; Send GAP3
	push eax
	call FDC_WriteByte

	mov eax,0FFh
	push eax
	call FDC_WriteByte	; DTL not used

	call FDC_WaitIRQ

	movzx eax,byte[status]
	and eax,0C0h
	or eax,eax
	jz @fdcrw_ok

	call FDC_Calibrate
	dec dword[temp]
	jz @fdcrw_ok
	jmp @fdcrw_loop
@fdcrw_ok:
	call FDC_MotorOff
@fdcrw_true:
	mov eax,1
	jmp @fdcrw_exit
@fdcrw_false:
	xor eax,eax
@fdcrw_exit:
	mov esp,ebp
	pop ebp
	ret 8

; Accepts a dword param but uses only the byte
FDC_WriteByte:
	push ebp
	mov ebp,esp

	xor eax,eax
	mov ecx,128
@fdwb_1:
	mov edx,FDC_MSR
	in al,dx
	and eax,0C0h
	cmp eax,80h
	jne @fdwb_2

	mov eax,[ebp+8]
	mov edx,FDC_DATA
	out dx,al
	jmp @fdwb_exit

@fdwb_2:
	dec ecx
	jnz @fdwb_1

	mov eax,-1
@fdwb_exit:
	mov esp,ebp
	pop ebp
	ret 4

FDC_ReadByte:
	push ebp
	mov ebp,esp

	xor eax,eax
	mov ecx,128
@fdrb_1:
	mov edx,FDC_MSR
	in al,dx
	and eax,0C0h
	cmp eax,0C0h
	jne @fdrb_2

	xor eax,eax
	mov edx,FDC_DATA
	in al,dx

	jmp @fdrb_exit
@fdrb_2:
	dec ecx
	jnz @fdrb_1

	mov eax,-1
@fdrb_exit:
	mov esp,ebp
	pop ebp
	ret

FDC_IRQ_Done	dd	0
FDC_MotorOnFlag	dd	0
FDC_Motor_Timer	dd	0
FDC_Timer		dd	0

status	times	7 db 0
temp		dd	0

sr0		db	0
fdc_track	db	0FFh